from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import map, zip
from future.utils import PY2, viewkeys, viewitems

import re
from itertools import chain
from locale import strxfrm
from functools import partial
from typing import TYPE_CHECKING

from .iter import switched_enumerate
from .binary import encode_binary, decode_binary

if TYPE_CHECKING:
	from typing import Any, Dict, Iterable, Tuple, Callable, Optional, Sequence, TypeVar, Union
	T = TypeVar("T")

english_consonants = "bcdfghjklmnpqrstvwxz" # y?
english_vowels = "aeiouy" # y?

german_consonants = "bcdfghjklmnpqrstvwxyz"
german_vowels = "aeiou"

#def cycleleft(s, start=0, end=-1): # bad
#	return s[:start] + s[start+1:end] + s[start] + s[end+1:]

#def cycleright(s, start=0, end=-1): # bad
#	return s[:start] + s[end] + s[start:end] + s[end+1:]

def toint(obj, default=None):
	# type: (Any, Optional[T]) -> Union[int, Optional[T]]

	try:
		return int(obj)
	except (ValueError, TypeError):
		return default

# was: startcut
def removeprefix(s, prefix):
	# type: (str, str) -> str

	""" If string `s` starts with `pre` cut it off and return the remaining string. """

	if s.startswith(prefix):
		return s[len(prefix):]
	else:
		return s

# was: endcut
def removesuffix(s, suffix):
	# type: (str, str) -> str

	""" If string `s` ends with `suffix` cut it off and return the preceding string. """

	if s.endswith(suffix):
		return s[:-len(suffix)]
	else:
		return s

def encode_case(s):
	# type: (str, ) -> bytes

	""" Encode case information of a string to a bit-encoded bytes string.
	"""

	return encode_binary(c.isupper() for c in s)

def decode_case(s, key):
	# type: (str, bytes) -> str

	""" Apply the case information encoded in `key` to string `s`.
	"""

	return "".join(c.upper() if b else c for c, b in zip(s, decode_binary(key)))

def tryint(obj):
	# type: (Any, ) -> Any

	""" Converts `obj` to int if possible and returns the original object otherwise.
	"""

	try:
		return int(obj)
	except (ValueError, TypeError):
		return obj

def locale_sorted(seq, case_insensitive=True, lower_before_upper=True):
	if case_insensitive:
		if lower_before_upper:
			key = lambda s: (strxfrm(s.lower()), strxfrm(s.swapcase()))
		else:
			key = lambda s: (strxfrm(s.lower()), strxfrm(s))
	else:
		if lower_before_upper:
			key = lambda s: (strxfrm(s.swapcase()), strxfrm(s))
		else:
			key = strxfrm

	return sorted(seq, key=key)

def build_multiple_replace(d, escape=True):
	# type: (Dict[str, str], bool) -> Callable

	if escape:
		it = map(re.escape, viewkeys(d))
	else:
		it = viewkeys(d)

	cp = re.compile("(" + "|".join(it) + ")")
	return partial(cp.sub, lambda m: d[m.group(0)])

def replace_multiple(s, d): # cython candidate...
	# type: (Iterable[str], Dict[str, str]) -> str

	""" Use dictionary d to replace instances of key with value.
		If input is a string, the keys must be strings of length 1.
		If the input is an iterable of strings,
		the strings will be compared to the keys for equality
	"""

	return "".join(d.get(c, c) for c in s)

def are_parentheses_matched(s, open="([{", close=")]}"):
	# type: (str, str, str) -> bool

	stack = list()
	parentheses = dict(chain(switched_enumerate(open), switched_enumerate(close)))
	assert len(parentheses) == len(open) + len(close)

	for char in s:
		if char in open:
			stack.append(parentheses[char])
		if char in close:
			try:
				if stack.pop() != parentheses[char]: # wrong nesting
					return False
			except IndexError: # stack empty
				return False

	return len(stack) == 0 # unclosed parentheses left

# was: cond_join
def filter_join(s, it, func=None):
	# type: (str, Iterable[str], Optional[Callable]) -> str

	if func is None:
		return s.join(i for i in it if i)
	else:
		return s.join(i for i in it if func(i))

def surrounding_join(j, it, l="", r=""):
	# type: (str, Iterable[str], str, str) -> str

	""" Example: surrounding_join(", ", ("a", "b", "c"), l="[", r="]") -> "[a], [b], [c]"
	"""

	s = (r+j+l).join(it)
	if s != "":
		return l+s+r
	return ""

def replace_pairs(s, items):
	# type: (str, Iterable[Tuple[str, str]]) -> str

	for key, value in items:
		s = s.replace(key, value)
	return s

def replace_pairs_bytes(s, items):
	# type: (bytes, Dict[bytes, Optional[bytes]]) -> bytes

	frm = b"".join(k for k, v in viewitems(items) if v)
	to = b"".join(v for k, v in viewitems(items) if v)
	delete = b"".join(k for k, v in viewitems(items) if v is None)

	if PY2:
		import string
		table = string.maketrans(frm, to)
	else:
		table = s.maketrans(frm, to)

	return s.translate(table, delete)

def replace_pairs_chars(s, items):
	# type: (str, Dict[str, Optional[str]]) -> str

	if PY2:
		# table = s.maketrans(items) # 'unicode' object has no attribute 'maketrans'
		table = {ord(k):v for k, v in viewitems(items)}
	else:
		table = s.maketrans(items)

	return s.translate(table)

def replace_list(s, character_list, replace_char):
	# type: (str, Sequence[str], str) -> str

	""" should work with CaseInsensitiveString also and replace replace_list_insensitive"""

	for find_char in character_list:
		s = s.replace(find_char, replace_char)
	return s



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import range
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Iterable, Iterator

def encode_binary(boolit, pad='0'):
	# type: (Iterable[bool], ) -> bytes

	bin = "".join('1' if b else '0' for b in boolit)
	bin += pad*(8 - len(bin)%8) # pad
	assert len(bin) % 8 == 0
	return bytes(int(bin[x:x+8], 2) for x in range(0, len(bin), 8))

def decode_binary(key):
	# type: (bytes, ) -> Iterator[bool]

	for b in key:
		for c in "{:08b}".format(b):
			yield True if c == '1' else False



from __future__ import absolute_import, division, print_function, unicode_literals

import pickle  # nosec
from functools import wraps
from datetime import timedelta
from typing import TYPE_CHECKING

from .atomic import TransactionalCreateFile
from .compat import FileNotFoundError
from .datetime import now
from .file import copen
from .filesystem import mdatetime
from .object import args_to_key

if TYPE_CHECKING:
	from typing import Any, Callable, Optional, Iterator, Iterable
	from .compat.pathlib import Path

def read_pickle(path):
	# type: (str, ) -> Any

	""" Read pickle file from `path`.
		Warning: All usual security consideration regarding the pickle module still apply.
	"""

	with copen(path, "rb") as fr:
		return pickle.load(fr)  # nosec

def write_pickle(result, path, protocol=None, safe=False):
	# type: (Any, str, Optional[int], bool) -> None

	""" Write `result` to `path` using pickle serialization.

		`protocol': pickle protocol version
		`safe`: if True, don't overwrite original file in case any error occurs
	"""

	if safe:
		context = TransactionalCreateFile
	else:
		context = copen

	with context(path, "wb") as fw:
		pickle.dump(result, fw, protocol=protocol)

def read_iter(path):
	# type: (str, ) -> Iterator[Any]

	""" Read pickled iterable from `path`. 
		Warning: All usual security consideration regarding the pickle module still apply.
	"""

	with copen(path, "rb") as fr:
		unpickler = pickle.Unpickler(fr)  # nosec
		while fr.peek(1):
			yield unpickler.load()

def write_iter(it, path, protocol=None, safe=False):
	# type: (Iterable[Any], str, Optional[int], bool) -> Iterator[Any]

	""" Write iterable `it` to `path` using pickle serialization. This uses much less memory than
			writing a full list at once.
		Read back using `read_iter()`. If `safe` is True, the original file is not overwritten
			if any error occurs.
		This is a generator which yields the values read from `it`. So it must be consumed
			to actually write anything to disk.
	"""

	if safe:
		context = TransactionalCreateFile
	else:
		context = copen

	with context(path, "wb") as fw:
		pickler = pickle.Pickler(fw, protocol=protocol)
		for result in it:
			pickler.dump(result)
			yield result

def key_to_hash(key, protocol=None):
	from hashlib import md5
	binary = pickle.dumps(key, protocol=protocol)
	return md5(binary).hexdigest()  # nosec

def cache(path, duration=None, generator=False, protocol=None):
	# type: (Path, Optional[timedelta], bool, Optional[int]) -> Callable[[Callable], Callable]

	""" Decorator to cache function calls. Doesn't take function arguments into regard.
		It's using `pickle` to deserialize the data. So don't use it with untrusted inputs.

		`path`: path to cache file
		`duration`: maximum age of cache
		`generator`: set to True to store the results of generator objects
		`protocol`: pickle protocol version
	"""

	duration = duration or timedelta.max

	def decorator(func):
		# type: (Callable, ) -> Callable

		@wraps(func)
		def inner(*args, **kwargs):
			# type: (*Any, **Any) -> Any

			hash = key_to_hash(args_to_key(args, kwargs), protocol=protocol)
			fullpath = path / hash

			try:
				invalid = now() - mdatetime(fullpath) > duration
			except FileNotFoundError:
				invalid = True

			if invalid:
				path.mkdir(parents=True, exist_ok=True)
				if generator:
					it = func(*args, **kwargs)
					return write_iter(it, fullpath, protocol=protocol, safe=True)
				else:
					result = func(*args, **kwargs)
					write_pickle(result, fullpath, protocol=protocol, safe=True)
					return result
			else:
				if generator:
					return read_iter(fullpath)
				else:
					return read_pickle(fullpath)
		return inner
	return decorator



from __future__ import absolute_import, division, print_function, unicode_literals

import re

from .url import valid_uri_characters
from .string import replace_multiple

def extract_urls(text):
	pattern = r"https?\:\/\/[" + re.escape(valid_uri_characters) + r"]+"
	for match in re.finditer(pattern, text):
		url = match.group(0)
		yield url

def replace_typographical_punctuation(s):
	# type: (str, ) -> str

	""" Replaces typographical punctuation with ASCII punctuation. """

	d = {
		"‚": "'",
		"„": '"',
		"‘": "'",
		"’": "'",
		"“": '"',
		"”": '"',
		"‒": "-", # figure dash
		"–": "-", # en dash
		"—": "--", # em dash
		"―": "--", # horizontal bar
		"‗": "_", # double low line
		"…": "...", # horizontal ellipsis
	}
	return replace_multiple(s, d)
